<?php
namespace Extension\Templavoila\Utility;

/*
 * This file is part of the TYPO3 CMS project.
 *
 * It is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License, either version 2
 * of the License, or any later version.
 *
 * For the full copyright and license information, please read the
 * LICENSE.txt file that was distributed with this source code.
 *
 * The TYPO3 project - inspiring people to share!
 */

/**
 * Class with static functions for templavoila.
 *
 * @author Steffen Kamper  <info@sk-typo3.de>
 */
final class GeneralUtility {

	/**
	 * @return \TYPO3\CMS\Core\Database\DatabaseConnection
	 */
	static public function getDatabaseConnection() {
		return $GLOBALS['TYPO3_DB'];
	}

	/**
	 * @return \TYPO3\CMS\Core\Authentication\BackendUserAuthentication
	 */
	static public function getBackendUser() {
		return $GLOBALS['BE_USER'];
	}

	/**
	 * @return \TYPO3\CMS\Lang\LanguageService
	 */
	static public function getLanguageService() {
		return $GLOBALS['LANG'];
	}

	/**
	 * @return array
	 */
	static public function getDenyListForUser() {
		$denyItems = array();
		foreach (static::getBackendUser()->userGroups as $group) {
			$groupDenyItems = \TYPO3\CMS\Core\Utility\GeneralUtility::trimExplode(',', $group['tx_templavoila_access'], TRUE);
			$denyItems = array_merge($denyItems, $groupDenyItems);
		}

		return $denyItems;
	}

	/**
	 * Get a list of referencing elements other than the given pid.
	 *
	 * @param array $element array with tablename and uid for a element
	 * @param integer $pid the suppoed source-pid
	 * @param integer $recursion recursion limiter
	 * @param array &$references array containing a list of the actual references
	 *
	 * @return boolean true if there are other references for this element
	 */
	static public function getElementForeignReferences($element, $pid, $recursion = 99, &$references = NULL) {
		if (!$recursion) {
			return FALSE;
		}
		if (!is_array($references)) {
			$references = array();
		}
		$refrows = static::getDatabaseConnection()->exec_SELECTgetRows(
			'*',
			'sys_refindex',
			'ref_table=' . static::getDatabaseConnection()->fullQuoteStr($element['table'], 'sys_refindex') .
			' AND ref_uid=' . (int)$element['uid'] .
			' AND deleted=0'
		);

		if (is_array($refrows)) {
			foreach ($refrows as $ref) {
				if (strcmp($ref['tablename'], 'pages') === 0) {
					$references[$ref['tablename']][$ref['recuid']] = TRUE;
				} else {
					if (!isset($references[$ref['tablename']][$ref['recuid']])) {
						// initialize with false to avoid recursion without affecting inner OR combinations
						$references[$ref['tablename']][$ref['recuid']] = FALSE;
						$references[$ref['tablename']][$ref['recuid']] = self::hasElementForeignReferences(array('table' => $ref['tablename'], 'uid' => $ref['recuid']), $pid, $recursion - 1, $references);
					}
				}
			}
		}

		unset($references['pages'][$pid]);

		return $references;
	}

	/**
	 * Checks if a element is referenced from other pages / elements on other pages than his own.
	 *
	 * @param array $element array with tablename and uid for a element
	 * @param integer $pid the suppoed source-pid
	 * @param integer $recursion recursion limiter
	 * @param array &$references array containing a list of the actual references
	 *
	 * @return boolean true if there are other references for this element
	 */
	static public function hasElementForeignReferences($element, $pid, $recursion = 99, &$references = NULL) {
		$references = self::getElementForeignReferences($element, $pid, $recursion, $references);
		$foreignRefs = FALSE;
		if (is_array($references)) {
			unset($references['pages'][$pid]);
			$foreignRefs = count($references['pages']) || count($references['pages_language_overlay']);
		}

		return $foreignRefs;
	}

}
